/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

#ifndef lint
char copyright[] =
"@(#) Copyright (c) 1983 Regents of the University of California.\n\
 All rights reserved.\n";
#endif /* not lint */

#ifndef lint
static char sccsid[] = "@(#)inetd.c	5.14 (Berkeley) 1/23/89";
#endif /* not lint */

/*
 * Inetd - Internet super-server
 *
 * This program invokes all internet services as needed.
 * connection-oriented services are invoked each time a
 * connection is made, by creating a process.  This process
 * is passed the connection as file descriptor 0 and is
 * expected to do a getpeername to find out the source host
 * and port.
 *
 * Datagram oriented services are invoked when a datagram
 * arrives; a process is created and passed a pending message
 * on file descriptor 0.  Datagram servers may either connect
 * to their peer, freeing up the original socket for inetd
 * to receive further messages on, or ``take over the socket'',
 * processing all arriving datagrams and, eventually, timing
 * out.	 The first type of server is said to be ``multi-threaded'';
 * the second type of server ``single-threaded''.
 *
 * Inetd uses a configuration file which is read at startup
 * and, possibly, at some later time in response to a hangup signal.
 * The configuration file is ``free format'' with fields given in the
 * order shown below.  Continuation lines for an entry must begin with
 * a space or tab.  All fields must be present in each entry.
 *
 *	service name			must be in /etc/services or must
 *					name a tcpmux or rpc service
 *	socket type			stream/dgram/raw/rdm/seqpacket
 *	protocol			must be in /etc/protocols unless
 *					rpc service
 *	wait/nowait			single-threaded/multi-threaded
 *	user				user to run daemon as
 *	server program			full path name
 *	server program arguments	maximum of MAXARGS (10+1)
 *
 * TCP services without official port numbers are handled with the
 * RFC1078-based tcpmux internal service. Tcpmux listens on port 1 for
 * requests. When a connection is made from a foreign host, the service
 * requested is passed to tcpmux, which looks it up in the servtab list
 * and returns the proper entry for the service. Tcpmux returns a
 * negative reply if the service doesn't exist, otherwise the invoked
 * server is expected to return the positive reply if the service type in
 * inetd.conf file has the prefix "tcpmux/". If the service type has the
 * prefix "tcpmux/+", tcpmux will return the positive reply for the
 * process; this is for compatibility with older server code, and also
 * allows you to invoke programs that use stdin/stdout without putting any
 * special server code in them. Services that use tcpmux are "nowait"
 * because they do not have a well-known port and hence cannot listen
 * for new requests.
 *
 * Services based on RPC also constitute a special case.  These services
 * need not be listed in /etc/services, because their port numbers are
 * dynamically bound by the portmapper.  The portmapper protocol specifies
 * a service with the triple (program number, version min, version max).
 * The format for RPC service entries begins as follows:
 *
 *	service-name/version-set	service-name must be in /etc/rpc
 *					version-set has the form 1-3,7,10
 *	socket type			dgram or stream
 *	rpc/protocol			protocol is udp or tcp
 *	etc.				as above
 *
 * Comment lines are indicated by a `#' in column 1.
 *
 * Inetd normally logs an error to the syslog if a server program
 * is missing.  Placing a `?' at the beginning of the server program 
 * path (for example, ?/usr/etc/podd) will suppress the error message.
 *
 * The following options are supported when inetd runs on a kernel that
 * supports IP Security Options, and silently ignored otherwise: 
 * 
 * The string "/lc" (label-cognizant) after "wait" or "nowait" causes
 * inetd to invoke the server at the default process label and to pass a
 * wildcard socket to the server. Otherwise, the server is invoked with a
 * socket and process label corresponding with that of the client.
 * 
 * The string "/rcv" after the user name causes inetd to invoke the server
 * with the uid of the client.  This is useful for servers that
 * communicate with Xsgi.
 */ 

#define _BSD_SIGNALS

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/file.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <rpc/rpc.h>		/* includes <netinet/in.h> */
#include <rpc/pmap_prot.h>
#include <arpa/inet.h>

#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <netdb.h>
#include <syslog.h>
#include <pwd.h>
#include <stdio.h>
#include <strings.h>
#include <unistd.h>
#include <sys/cipso.h>
#include <net/soioctl.h>
#include <sys/mac_label.h>
#include <clearance.h>
#include <mls.h>
#include <sys/types.h>
#include <sys/so_dac.h>

#define	TOOMANY		1000		/* don't start more than TOOMANY */
#define	LISTEN_QLEN	255		/* Maximum length of list queue */
#define	CNT_INTVL	60		/* servers in CNT_INTVL sec. */
#define	RETRYTIME	(60*10)		/* retry after bind or server fail */

#define	SIGBLOCK	(sigmask(SIGCHLD)|sigmask(SIGHUP)|sigmask(SIGALRM))

void	config(), reapchild(), retry();

char	*emalloc(int);
#define	tmalloc(type)	((type *) emalloc(sizeof(type)))

int	debug = 0;
int	nsock, maxsock;
fd_set	allsock;
int	options;
int	timingout;
int 	toomany = TOOMANY;

struct rpcvers {
	u_long		rv_low;		/* low version number */
	u_long		rv_high;	/* same as low or high if range */
	struct rpcvers	*rv_more;	/* more numbers or ranges */
};

struct	servtab {
	char	*se_service;		/* name of service */
	int	se_socktype;		/* type of socket to use */
	char	*se_proto;		/* protocol used */
	short	se_wait;		/* single threaded server */
	short	se_checked;		/* looked at during merge */
	char	*se_user;		/* user name to run as */
	struct	biltin *se_bi;		/* if built-in, description */
	char	*se_server;		/* server program */
	short   se_pid;                 /* daemon process id */
#define MAXARGV 11
	char	*se_argv[MAXARGV+1];	/* program arguments */
	int	se_fd;			/* open descriptor */
	int	se_type;		/* discriminant for se_un */
	union {
		struct	sockaddr_in ctrladdr;
		struct {
			u_long prog;	/* RPC program number */
			u_short prot;	/* IP protocol number */
			u_short port;	/* port in net byte order */
			struct rpcvers vers;
		} rpcinfo;
	} se_un;
	int	se_count;		/* number started since se_time */
	struct	timeval se_time;	/* start of se_count */
	struct	servtab *se_aliases;	/* head of alias list */
	struct	servtab *se_nextalias;	/* alias forward linkage */
	struct	servtab *se_next;
	int	se_flags;
} *servtab;

#define se_ctrladdr	se_un.ctrladdr
#define se_rpcinfo	se_un.rpcinfo
#define	se_prog		se_rpcinfo.prog
#define	se_prot		se_rpcinfo.prot
#define	se_port		se_rpcinfo.port
#define	se_vers		se_rpcinfo.vers

#define NORM_TYPE	0
#define RPC_TYPE	1
#define MUX_TYPE	2
#define MUXPLUS_TYPE	3
#define ISRPC(sep)	((sep)->se_type == RPC_TYPE)
#define ISMUX(sep)	(((sep)->se_type == MUX_TYPE) || \
			 ((sep)->se_type == MUXPLUS_TYPE))
#define ISMUXPLUS(sep)	((sep)->se_type == MUXPLUS_TYPE)

/* se_flags: */
#define	SE_MISSING_OK	0x00000001	/* Don't log error if missing */
#define IS_MISSING_OK(sep) ((sep)->se_flags & SE_MISSING_OK)
#define	SE_LBLCOG	0x00000002	/* Leave socket label wildcard */
#define IS_LBLCOG(sep) ((sep)->se_flags & SE_LBLCOG)
#define	SE_CHAMELEON	0x00000004	/* Set chameleon socket option */
#define IS_CHAMELEON(sep) ((sep)->se_flags & SE_CHAMELEON)
#define	SE_RCVUID	0x00000008	/* Use received (client) uid */
#define IS_RCVUID(sep) ((sep)->se_flags & SE_RCVUID)

int echo_stream(), discard_stream(), machtime_stream();
int daytime_stream(), chargen_stream();
int echo_dg(), discard_dg(), machtime_dg(), daytime_dg(), chargen_dg();
struct servtab *tcpmux();

struct biltin {
	char	*bi_service;		/* internally provided service name */
	int	bi_socktype;		/* type of socket supported */
	short	bi_fork;		/* 1 if should fork before call */
	short	bi_wait;		/* 1 if should wait for child */
	int	(*bi_fn)();		/* function which performs it */
} biltins[] = {
	/* Echo received data */
	"echo",		SOCK_STREAM,	1, 0,	echo_stream,
	"echo",		SOCK_DGRAM,	0, 0,	echo_dg,

	/* Internet /dev/null */
	"discard",	SOCK_STREAM,	1, 0,	discard_stream,
	"discard",	SOCK_DGRAM,	0, 0,	discard_dg,

	/* Return 32 bit time since 1970 */
	"time",		SOCK_STREAM,	0, 0,	machtime_stream,
	"time",		SOCK_DGRAM,	0, 0,	machtime_dg,

	/* Return human-readable time */
	"daytime",	SOCK_STREAM,	0, 0,	daytime_stream,
	"daytime",	SOCK_DGRAM,	0, 0,	daytime_dg,

	/* Familiar character generator */
	"chargen",	SOCK_STREAM,	1, 0,	chargen_stream,
	"chargen",	SOCK_DGRAM,	0, 0,	chargen_dg,

	"tcpmux",	SOCK_STREAM,	1, 0,	(int (*)())tcpmux,

	0
};

static short havemac = 0;   	/* kernel supports Mandatory Access Control */
static short havecipso = 0; 	/* kernel supports IP secopts & socket ACL */
mac_label original_label;
mac_label equal_label = { MSEN_EQUAL_LABEL, MINT_EQUAL_LABEL, 0, 0, 0, 0 };
void check_clearance(struct clearance *, struct servtab *sep);
static char rationale[256];	/* for ia_audit */

#define NUMINT	(sizeof(intab) / sizeof(struct inent))
char	*CONFIG = "/etc/inetd.conf";

extern char **environ;
static char timez[256] = { "TZ=" };
static char *env[] = {
	timez, NULL,
};

static char *
mystrdup(const char *cp)
{
	if (cp == NULL)
		cp = "";
	return strcpy(emalloc(strlen(cp) + 1), cp);
}

main(argc, argv)
	int argc;
	char *argv[];
{
	extern char *optarg;
	extern int optind, opterr;
	register struct servtab *sep, *next;
	register int tmp;
	struct sigvec sv;
	int opt, pid, dofork;
	char buf[50];
	char *tp;
	struct soacl sacl;

	openlog("inetd", LOG_PID | LOG_NOWAIT, LOG_DAEMON);
	if (getuid()) {
		fprintf(stderr, "inetd: must be started by super-user\n");
		syslog(LOG_ERR, "usage: inetd must be started by super-user.");
		exit(1);
	}

	havemac = (sysconf(_SC_MAC) == 1);
	havecipso = (sysconf(_SC_IP_SECOPTS) == 1);
	if (havecipso) {
		/* 
		 * Set WILDACL process socket acl before creating any sockets. 
		 */	
		bzero(&sacl, sizeof sacl);
		sacl.so_uidcount = WILDACL; 
		if (sgi_setpsoacl(&sacl) < 0) {
			syslog(LOG_ERR, "can't set psoacl: %m");
			_exit(1);
		}
	}
	opterr = 0;
	while ((opt = getopt(argc, argv, "dR:")) != EOF)
		switch (opt) {
		case 'd':
			debug = 1;
#if 0
			options |= SO_DEBUG;
#endif
			break;
		case 'R':	/* invocation rate */
			tmp = atoi(optarg);
			if (tmp < 1)
				syslog(LOG_ERR,
			         "-R %d: bad value for service invocation rate",
					tmp);
			else
				toomany = tmp;
			break;
		default:
			syslog(LOG_ERR, "usage: inetd [-R rate] [conf-file]");
			exit(1);
		}
	argc -= optind;
	argv += optind;

	if (argc > 0)
		CONFIG = argv[0];

	/* disable debugging if invoked from startup script */
	if (debug && !_isatty(0))
		debug = 0;
	if (debug == 0) {
		if (_daemonize(0, -1, -1, -1) < 0)
			exit(1);
		/* _daemonize does a closelog() */
		openlog("inetd", LOG_PID | LOG_NOWAIT, LOG_DAEMON);
	}

	if ((tp = getenv("TZ")) != NULL)
		strncat(timez,tp,sizeof(timez)-3);
	else
		env[0] = NULL;
	environ = env;
	if (havemac && havecipso) {
		if (sgi_getplabel(&original_label) < 0) {
			syslog(LOG_ERR, "%s: can't get plabel: %m",
			       sep->se_service);
			exit(1);
		}
	}
	bzero((char *)&sv, sizeof(sv));
	sv.sv_mask = SIGBLOCK;
	sv.sv_handler = retry;
	sigvec(SIGALRM, &sv, (struct sigvec *)0);
	config();
	sv.sv_handler = config;
	sigvec(SIGHUP, &sv, (struct sigvec *)0);
	sv.sv_handler = reapchild;
	sigvec(SIGCHLD, &sv, (struct sigvec *)0);

	for (;;) {
	    int n, ctrl;
	    fd_set readable;

	    if (nsock == 0) {
		(void) sigblock(SIGBLOCK);
		while (nsock == 0)
		    sigpause(0L);
		(void) sigsetmask(0L);
	    }
	    readable = allsock;
	    if ((n = select(maxsock + 1, &readable, (fd_set *)0,
		(fd_set *)0, (struct timeval *)0)) <= 0) {
		    if (n < 0 && errno != EINTR){
			syslog(LOG_WARNING, "select: %m\n");
                        sginap(HZ);
                    }
                    else {
                        /* Came here due to a signal while waiting on select
                         * Most often, it's due to SIGCHLD - Child process
                         * started by inetd created another one to do the
                         * work and dies. Allows inetd to start next instance
                         * of a "wait" type server. If inetd sleeps for
                         * a second in each such case, we end up allowing
                         * these servers to start at max rate of 1/sec.
                         * So sginap(1) should improve this rate.. But
                         * what about child death due to some server being
                         * removed ?? -> Let '-R' handle it.
                         */
                        sginap(1);
                    }
		    continue;
	    }
	    for (sep = servtab; n && sep; sep = next) {
		struct servtab *alias;
		mac_label socket_label;
		uid_t socket_uid = -2;		/* make default uid invalid */
		struct recvluinfo rluinfo;
		char *usename;
		int cc;				/* buf for recvlu */

		rluinfo.rlu_label = &socket_label;
		rluinfo.rlu_uidp = &socket_uid;

		if (havecipso && havemac) /* make default label invalid */
			socket_label.ml_msen_type = MSEN_UNKNOWN_LABEL;

		next = sep->se_next;
		if (sep->se_fd < 0 || !FD_ISSET(sep->se_fd, &readable))
			continue;
		n--;
		if (debug)
			fprintf(stderr, "someone wants %s on fd %d\n", 
				sep->se_service, sep->se_fd);
		if (sep->se_socktype == SOCK_STREAM) {
			if (havecipso) {
				ctrl = sep->se_fd;
				if ((ISRPC(sep) || sep->se_wait) && IS_CHAMELEON(sep)) {
				/* STREAM RPC or wait, chameleon */
					if (ioctl(ctrl, SIOCGETLABEL, &socket_label) < 0) {
					     syslog(LOG_ERR, 
						    "%s/%s: SIOCGETLABEL ioctl: %m",
						    sep->se_service, sep->se_proto);
						(void)close(sep->se_fd);
						FD_CLR(sep->se_fd, &allsock);
						nsock--;
						sep->se_fd = -1;
						continue;
					} 
					if (ioctl(ctrl, SIOCGETRCVUID, &socket_uid) < 0) {
					     syslog(LOG_ERR, 
						    "%s/%s: SIOCGETRCVUID ioctl: %m",
						    sep->se_service, sep->se_proto);
						(void)close(sep->se_fd);
						FD_CLR(sep->se_fd, &allsock);
						nsock--;
						sep->se_fd = -1;
						continue;
					} 
				}
				if (!ISRPC(sep) && !sep->se_wait) { 
					/* STREAM, non-RPC, nowait */
					ctrl = accept(sep->se_fd, (struct sockaddr *)0, (int *)0);
					if (debug)
						fprintf(stderr, "accept, ctrl %d\n", ctrl);
					if (ctrl < 0) {
						if (errno != EINTR)
							syslog(LOG_WARNING, "accept: %m");
						continue;
					}
					if (recvlu(ctrl, &cc, 0,
						   MSG_PEEK, &rluinfo) < 0)  {
						syslog(LOG_ERR, 
						       "%s/%s: recvlu failed: %m",
						       sep->se_service, sep->se_proto);
						(void)close(ctrl);
						continue;
					}
				}
			} else if (!sep->se_wait) {
				ctrl = accept(sep->se_fd, (struct sockaddr *)0, (int *)0);
				if (debug)
					fprintf(stderr, "accept, ctrl %d\n", ctrl);
				if (ctrl < 0) {
					if (errno != EINTR)
						syslog(LOG_WARNING, "accept: %m");
					continue;
				}
				/*
				 * Call tcpmux to find the real service to exec.
				 */
				if (sep->se_bi &&
				    sep->se_bi->bi_fn == (int (*)()) tcpmux) {
					sep = tcpmux(ctrl);
					if (sep == NULL) {
						close(ctrl);
						continue;
					}
				}
			}
			else /* SOCK_STREAM, wait, not cipso */
				ctrl = sep->se_fd;
		} else { 	/* SOCK_DGRAM */
			ctrl = sep->se_fd;
			if (havecipso) { 	/* get attributes */
				if (recvlu(ctrl, &cc, 0, MSG_PEEK, 
					   &rluinfo) < 0) {
					syslog(LOG_ERR, "%s/%s: recvlu failed: %m",
					       sep->se_service, sep->se_proto);
					
					FD_CLR(alias->se_fd,&allsock);
					nsock--;
					(void)close(sep->se_fd);
					FD_CLR(sep->se_fd, &allsock);
					nsock--;
					sep->se_fd = -1;
					continue;
				}
			}
		}
		(void) sigblock(SIGBLOCK);
		pid = 0;
		dofork = (sep->se_bi == 0 || sep->se_bi->bi_fork);
		if (dofork) {
			if (sep->se_count++ == 0)
				(void) gettimeofday(&sep->se_time, NULL);
			else if (sep->se_count >= toomany) {
				struct timeval now;

				(void) gettimeofday(&now, NULL);
				if (now.tv_sec - sep->se_time.tv_sec >
				    CNT_INTVL) {
					sep->se_time = now;
					sep->se_count = 1;
				} else {
					syslog(LOG_ERR,
			"%s/%s server failing (looping), service terminated",
					    sep->se_service, sep->se_proto);
					close_sep(sep);
					(void)close(ctrl);
					sigsetmask(0L);
					if (!timingout) {
						timingout = 1;
						alarm(RETRYTIME);
					}
					continue;
				}
			}
			pid = fork();
		}
		if (pid < 0) {
			if ((!sep->se_wait && sep->se_socktype == SOCK_STREAM) 
				|| (havecipso && ctrl != sep->se_fd)) 
				close(ctrl);
			sigsetmask(0L);
			sleep(1);
			continue;
		}
		if (pid && sep->se_wait) {
			alias = sep->se_aliases;
			if (alias == NULL) {
			        sep->se_pid=pid;
				FD_CLR(sep->se_fd, &allsock);
				nsock--;
			} else {
				do {
					alias->se_pid = pid;
					FD_CLR(alias->se_fd, &allsock);
					nsock--;
					alias = alias->se_nextalias;
				} while (alias != NULL);
			}
		}
		sigsetmask(0L);
		if (pid == 0) {
			if (debug && dofork)
				if (setsid() < 0)
					syslog(LOG_ERR, "setsid failed: %m");
			/* 
			 * MAC: set the label unless this server
			 * is label-cognizant (e.g. mountd).
			 */
			if (havecipso && havemac && !IS_LBLCOG(sep) &&
			    sgi_setplabel(&socket_label) < 0) {
				syslog(LOG_ERR, "%s/%s: setplabel: %m",
				       sep->se_service, sep->se_proto);
				if (dofork) /* child */
					exit(1);
				if (ctrl != sep->se_fd)
					close(ctrl);
				continue;
			}

			if (sep->se_bi) {
				(*sep->se_bi->bi_fn)(ctrl, sep);
				if (havemac && havecipso) {  /* set it back */
					if (sgi_setplabel(&original_label) < 0) {
						syslog(LOG_ERR, "setplabel: %m");
						exit(1);
					}
				}
			} else { 
				int lastfd;
				struct clearance *clp;
				struct passwd *pwd;

				if (debug) {
					fprintf(stderr, "%d execl %s\n",
						getpid(), sep->se_server);
				}
				alias = sep->se_aliases;
				if (alias == NULL) {
					dup2(ctrl, 0);
					dup2(0, 1);
					dup2(0, 2);
					lastfd = 2;
				} else {
					for (lastfd = 0; ; lastfd++) {
						dupaliasfd(alias, lastfd);
						alias = alias->se_nextalias;
						if (alias == NULL)
							break;
					}
				}

				if (havecipso && IS_RCVUID(sep))
					/*
					 * This server will be run with
					 * client's uid. (Used to restrict
					 * dgld's X server access.)
					 */
					pwd = getpwuid(socket_uid);
				else
					pwd = getpwnam(sep->se_user);
				usename = pwd->pw_name;

				/* 
				 * Check user clearance; audit the event.
				 */
				if (havemac) {
					if (!(clp = sgi_getclearancebyname(
					    usename))) {
						syslog(LOG_ERR,
			    "%s/%s: sgi_getclearancebyname: %s: No such user",
						    sep->se_service,
						    sep->se_proto, usename);
						if (sep->se_socktype !=
						    SOCK_STREAM)
							recv(0, buf,
							    sizeof (buf), 0);
						_exit(1);
					}
					check_clearance(clp, sep);
				}
				/*
				 * Audit
				 */
				sprintf(rationale, "Remote %s service granted.", 
					sep->se_service);
				(void)ia_audit("inetd", usename, 1, rationale);
				if (pwd->pw_uid) {
					if (setgid(pwd->pw_gid) < 0) {
						syslog(LOG_ERR,
						 "%s: can't set gid %d: %m", 
						       sep->se_service, pwd->pw_gid);
						_exit(1);
					}
					(void) initgroups(usename, pwd->pw_gid);
					if (setuid(pwd->pw_uid) < 0) {
						syslog(LOG_ERR,
						       "%s: can't set uid %d: %m", 
						       sep->se_service, 
						       pwd->pw_uid);
						_exit(1);
					}
				}
				if (havecipso) {
					/* 
					 * DAC: Server's psoacl should allow 
					 * received uid and new user id.
					 */
					bzero(&sacl, sizeof sacl);
					if (pwd->pw_uid != socket_uid && socket_uid > 0) {
						sacl.so_uidcount = 2; 
						sacl.so_uidlist[0] = pwd->pw_uid; 
						sacl.so_uidlist[1] = socket_uid; 
					} else
						sacl.so_uidcount = 0;
					if (sgi_setpsoacl(&sacl) < 0) {
						syslog(LOG_ERR,
						       "%s: can't set psoacl: %m",
						       sep->se_service);
						_exit(1);
					}
				}

				/*
				 * Close all of the sockets opened by NIS to
				 * do the getpwnam()
				 */
				for (tmp = getdtablehi(); --tmp > lastfd; )
					(void)close(tmp);
				execv(sep->se_server, sep->se_argv);
				if (sep->se_socktype != SOCK_STREAM)
					recv(0, buf, sizeof (buf), 0);
				syslog(LOG_ERR,
				    "cannot execute %s: %m", sep->se_server);
				_exit(1);
			}
		}
		if (!havecipso || !havemac) {
			if (!sep->se_wait && sep->se_socktype == SOCK_STREAM)
				close(ctrl);
		} else {
			if (ctrl != sep->se_fd)	{
				(void)close(ctrl);
			}
			if (pid > 0 && IS_CHAMELEON(sep) ) {
				(void)close(sep->se_fd);
				FD_CLR(sep->se_fd, &allsock);
				nsock--;
				sep->se_fd = -1;
				if (ISRPC(sep)) 
					rpcsetup(sep);
				else
					setup(sep);
			}	
		}
	    }
	}
}

dupaliasfd(alias, tofd)
	struct servtab *alias;
	int tofd;
{
	struct servtab *sep;

	if (alias->se_fd == tofd)
		return;
	for (sep = alias->se_aliases; sep != NULL; sep = sep->se_nextalias)
		if (sep != alias && sep->se_fd == tofd) {
			sep->se_fd = dup(tofd);
			if (sep->se_fd < 0)
				syslog(LOG_ERR, "%s: %m", sep->se_server);
		}
	dup2(alias->se_fd, tofd);
	alias->se_fd = tofd;
}

void
reapchild()
{
	union wait status;
	int pid;
	register struct servtab *sep;

	for (;;) {
		pid = wait3((int *)&status, WNOHANG, (struct rusage *)0);
		if (pid <= 0)
			break;
		if (debug)
			fprintf(stderr, "%d reaped, status %#x\n", 
				pid, status.w_status);
		for (sep = servtab; sep; sep = sep->se_next)
			if (sep->se_pid == pid) {
				if (status.w_status)
					syslog(LOG_WARNING,
					    "%s: exit status 0x%x",
					    sep->se_server, status.w_status);
				if (debug)
					fprintf(stderr, "restored %s, fd %d\n",
					    sep->se_service, sep->se_fd);
				if (!havecipso || sep->se_fd >= 0) {
					FD_SET(sep->se_fd, &allsock);
					nsock++;
				}
				sep->se_pid=1;
			}
	}
}

/*
 * Remove a servtab entry from its lists of aliases, if it's on any.
 * We call this before calling freeconfig, so that we don't end up
 * having pointers to free memory on alias lists.
 */
void
removealias(struct servtab *alias)
{
	struct servtab *sp, **app;

	for (sp = servtab; sp; sp = sp->se_next) {
		app = &sp->se_aliases;

		while (*app) {
			if (*app == alias) {
				*app = (*app)->se_nextalias;
			}
			if (*app) {
				app = &(*app)->se_nextalias;
			}
		}
	}
}

FILE	*fconfig = NULL;

void
config()
{
	register struct servtab *sep, *cp, **sepp;
	struct servtab *getconfigent(), *enter();
	long omask;
	struct servtab *alias;
	struct servent *sp;
	int svccnt = 0;
#ifdef sgi
	char cbuffer[1024];
	extern int _pw_stayopen, _getpwent_no_shadow;
#endif

	if (!setconfig()) {
		syslog(LOG_ERR, "%s: %m", CONFIG);
		return;
	}
#ifdef sgi
	setbuffer(fconfig, cbuffer, sizeof cbuffer);
	_pw_stayopen = _getpwent_no_shadow = 1;
#endif
	for (sep = servtab; sep; sep = sep->se_next)
		sep->se_checked = 0;
	while (cp = getconfigent()) {
		/* skip non-existent servers */
		if (cp->se_bi == 0 && access(cp->se_server, EX_OK) < 0) {
			if (errno != ENOENT || !IS_MISSING_OK(cp))
				syslog(LOG_ERR, "%s: %m", cp->se_server);
			continue;
		}
		if  ( (getpwnam(cp->se_user) == NULL) && !IS_RCVUID(cp) ) {
			syslog(LOG_ERR,
				"%s/%s: No such user '%s', service ignored",
				cp->se_service, cp->se_proto, cp->se_user);
			continue;
		}
		alias = 0;
		for (sep = servtab; sep; sep = sep->se_next) {
			if (strcmp(sep->se_server, cp->se_server) != 0)
				continue;
			if (strcmp(sep->se_service, cp->se_service) == 0 &&
			    strcmp(sep->se_proto, cp->se_proto) == 0)
				break;
			if (sep->se_wait &&
			    strcmp(sep->se_user, cp->se_user) == 0)
				alias = sep;
		}
		if (sep != 0) {
			int i;

			omask = sigblock(SIGBLOCK);

			/*
			 * If the RPC service info has changed, close the
			 * existing socket.  Replace the version set with
			 * cp's set, which we hide from freeconfig.
			 */
			if (ISRPC(sep) && sep->se_fd >= 0 &&
			    (sep->se_prog != cp->se_prog ||
			     !eqrpcvers(&sep->se_vers, &cp->se_vers))) {
				close_sep(sep);
				freerpcvers(sep);
				sep->se_rpcinfo = cp->se_rpcinfo;
				cp->se_vers.rv_more = NULL;
			}

			/*
			 * sep->se_wait may be holding the process id
			 * of a running daemon we are awaiting.  If so,
			 * don't lose it unless we are instructed now
			 * not to wait.
			 */
			if (cp->se_bi == 0) {
				sep->se_wait = cp->se_wait;
			}

#define SWAP(a, b) { char *c = a; a = b; b = c; }
			if (cp->se_user)
				SWAP(sep->se_user, cp->se_user);
			if (cp->se_server)
				SWAP(sep->se_server, cp->se_server);
			for (i = 0; i < MAXARGV; i++)
				SWAP(sep->se_argv[i], cp->se_argv[i]);
			sigsetmask(omask);
			/*
			 * We don't have to call removealias here,
			 * because there's no way that cp can be on an
			 * alias list.  We just modified sep, which
			 * will stay on any list it's already on.
			 */
			freeconfig(cp);
			if (debug)
				print_service("REDO", sep);
		} else {
			sep = enter(cp);
			if (alias) {
			    /*
			     * Sort sep into the alias list in canonical
			     * service and protocol order (reuse cp).
			     */
			    if (alias->se_aliases == NULL)
				    alias->se_aliases = alias;
			    for (sepp = &alias->se_aliases;
				 (cp = *sepp) != NULL &&
				 (strcmp(cp->se_service,sep->se_service) < 0 ||
				  (strcmp(cp->se_service,sep->se_service) == 0
				   && strcmp(cp->se_proto,sep->se_proto) < 0));
				 sepp = &cp->se_nextalias)
				    continue;
			    sep->se_aliases = alias->se_aliases;
			    sep->se_nextalias = cp;
			    *sepp = sep;
			}
			if (debug)
				print_service("ADD ", sep);
		}
		sep->se_checked = 1;
		if (ISMUX(sep)) {
			sep->se_fd = -1;
			continue;
		}
		if (ISRPC(sep)) {
			/*
			 * Don't remap RPC services if they haven't changed.
			 */
			if (sep->se_fd < 0)
				rpcsetup(sep);
			continue;
		}
		sp = getservbyname(sep->se_service, sep->se_proto);
		if (sp == 0) {
			syslog(LOG_ERR, "%s/%s: unknown service",
			    sep->se_service, sep->se_proto);
			sep->se_checked = 0;
			continue;
		}
		if (sp->s_port != sep->se_ctrladdr.sin_port) {
			sep->se_ctrladdr.sin_port = sp->s_port;
			if (sep->se_fd >= 0)
				close_sep(sep);
		}
		if (sep->se_fd < 0)
			setup(sep);
	}
	endconfig();
#ifdef sgi
	_pw_stayopen = 0;
	endpwent();
#endif
	/*
	 * Purge anything not looked at above.
	 */
	omask = sigblock(SIGBLOCK);
	sepp = &servtab;
	while (sep = *sepp) {
		if (sep->se_checked) {
			sepp = &sep->se_next;
			svccnt++;
			continue;
		}
		*sepp = sep->se_next;
		if (sep->se_fd >= 0) {
			close_sep(sep);
		}
		if (debug)
			print_service("FREE", sep);
		removealias(sep);
		freeconfig(sep);
		free((char *)sep);
	}
	(void) sigsetmask(omask);
	/*
	 * Add 2 fd's for accept and getpwnam calls.
	 */
	if (svccnt+2 >= getdtablesize())
		syslog(LOG_ERR, "too many services: open-file limit reached.");
}

int
eqrpcvers(rv1, rv2)
	struct rpcvers *rv1, *rv2;
{
	if (rv1->rv_low == rv2->rv_low && rv1->rv_high == rv2->rv_high) {
		if (rv1->rv_more && rv2->rv_more)
			return eqrpcvers(rv1->rv_more, rv2->rv_more);
		if (rv1->rv_more || rv2->rv_more)
			return 0;
		return 1;
	}
	return 0;
}

void
retry()
{
	register struct servtab *sep;

	timingout = 0;
	for (sep = servtab; sep; sep = sep->se_next) {
		if (sep->se_fd < 0) {
			if (ISRPC(sep))
				rpcsetup(sep);
			else
				setup(sep);
		}
	}
}

int
turnon( int fd, int opt)
{
	int on = 1;

	on = setsockopt(fd, SOL_SOCKET, opt, (char *)&on, sizeof on);
	return on;
}

setup(sep)
	register struct servtab *sep;
{
	int on = 1;

	if ((sep->se_fd = socket(AF_INET, sep->se_socktype, 0)) < 0) {
		syslog(LOG_ERR, "%s/%s: socket: %m",
		    sep->se_service, sep->se_proto);
		return;
	}
	if (havecipso) {
		if (ioctl(sep->se_fd, SIOCSETLABEL, &equal_label ) < 0 ) {
			syslog(LOG_ERR, "%s/%s: SIOCSETLABEL ioctl: %m",
			    sep->se_service, sep->se_proto);
			(void)close(sep->se_fd);
			sep->se_fd = -1;
			return;
		}
		if (IS_CHAMELEON(sep)) {
			if (turnon(sep->se_fd, SO_CHAMELEON ) < 0) {
				syslog(LOG_ERR, 
					"%s/%s: SO_CHAMELEON setsockopt: %m",
					sep->se_service, sep->se_proto);
				(void)close(sep->se_fd);
				sep->se_fd = -1;
				return;
			}
			if (turnon(sep->se_fd, SO_REUSEPORT) < 0) {
				syslog(LOG_ERR, 
					"%s/%s: SO_REUSEPORT setsockopt: %m",
					sep->se_service, sep->se_proto);
				(void)close(sep->se_fd);
				sep->se_fd = -1;
				return;
			}
		}
	}
#define	turnon(fd, opt) \
setsockopt(fd, SOL_SOCKET, opt, (char *)&on, sizeof (on))
	if (strcmp(sep->se_proto, "tcp") == 0 && (options & SO_DEBUG) &&
	    turnon(sep->se_fd, SO_DEBUG) < 0)
		syslog(LOG_ERR, "setsockopt (SO_DEBUG): %m");
	if (turnon(sep->se_fd, SO_REUSEADDR) < 0)
		syslog(LOG_ERR, "setsockopt (SO_REUSEADDR): %m");
#undef turnon
	if (bind(sep->se_fd, &sep->se_ctrladdr,
	    sizeof (sep->se_ctrladdr)) < 0) {
		syslog(LOG_ERR, "%s/%s: bind: %m",
		    sep->se_service, sep->se_proto);
		(void) close(sep->se_fd);
		sep->se_fd = -1;
		if (!timingout) {
			timingout = 1;
			alarm(RETRYTIME);
		}
		return;
	}
	if (sep->se_socktype == SOCK_STREAM) 
		listen(sep->se_fd, IS_CHAMELEON(sep) ? 1 : LISTEN_QLEN);
	FD_SET(sep->se_fd, &allsock);
	nsock++;
	if (sep->se_fd > maxsock)
		maxsock = sep->se_fd;
	if (debug) {
		fprintf(stderr, "registered %s on %d\n",
			sep->se_server, sep->se_fd);
	}
}

/*
 * Finish with a service and its socket.
 */
close_sep(sep)
	register struct servtab *sep;
{
	if (sep->se_fd >= 0) {
		nsock--;
		FD_CLR(sep->se_fd, &allsock);
		(void) close(sep->se_fd);
		sep->se_fd = -1;
	}
	/*
	 * Don't keep the pid of this running deamon: when reapchild()
	 * reaps this pid, it would erroneously increment nsock.
	 */
	if (sep->se_pid > 1){
	  sep->se_pid=1;
	}
	sep->se_count = 0;
	if (ISRPC(sep)) {
		struct rpcvers *rvp;
		u_long vers;

		rvp = &sep->se_vers;
		do {
			for (vers = rvp->rv_low; vers <= rvp->rv_high; vers++)
				pmap_unset(sep->se_prog, vers);
			rvp = rvp->rv_more;
		} while (rvp);
		if (sep->se_aliases != NULL)
			pmap_reset_aliases(sep, 1);
	}
}

/*
 * Thanks to Sun's unorthogonal pmap_unset, which unsets all portmappings
 * for prog&vers, we must restore other protocols' mappings when unmapping
 * rpcsep's port.
 */
pmap_reset_aliases(rpcsep, skipit)
	struct servtab *rpcsep;
	int skipit;
{
	struct servtab *sep;
	struct rpcvers *rvp;
	u_long vers;

	for (sep = rpcsep->se_aliases; sep; sep = sep->se_nextalias) {
		/* Restore just the other services' mappings, not rcpsep's */
		if ((skipit && sep == rpcsep) || (sep->se_checked == 0))
			continue;

		rvp = &sep->se_vers;
		do {
			for (vers = rvp->rv_low; vers <= rvp->rv_high; vers++) {
				pmap_set(sep->se_prog, vers, sep->se_prot,
					 sep->se_port);
			}
			rvp = rvp->rv_more;
		} while (rvp);
	}
}

rpcsetup(sep)
	register struct servtab *sep;
{
	register int rpcsock;

	if (sep->se_fd >= 0)
		close_sep(sep);
	rpcsock = getrpcsock(sep);
	if (rpcsock < 0) {
		syslog(LOG_ERR, "rpc/%s socket creation problem: %m",
		    sep->se_proto);
		sep->se_checked = 0;
		return;
	}
	sep->se_fd = rpcsock;
	FD_SET(rpcsock, &allsock);
	nsock++;
	if (rpcsock > maxsock)
		maxsock = rpcsock;
	if (debug) {
		fprintf(stderr, "registered %s on %d\n",
			sep->se_server, rpcsock);
	}
}

int
getrpcsock(sep)
	struct servtab *sep;
{
	int s, len, listen_qlen;
	struct sockaddr_in addr;
	struct rpcvers *rvp;
	int registrations = 0;
	int port;

	if ((s = socket(AF_INET, sep->se_socktype, 0)) < 0) {
			syslog(LOG_ERR, "socket create failed: type %d",
				sep->se_socktype);
		return -1;
	}
	if (havecipso && havemac) {
		if (ioctl(s, SIOCSETLABEL, &equal_label ) < 0 ) {
			syslog(LOG_ERR, "%s/%s: rpc SIOCSETLABEL ioctl: %m",
		    	sep->se_service, sep->se_proto);
			(void)close(s);
			return -1;
		}
		if (IS_CHAMELEON(sep)) {
				if (turnon(s, SO_CHAMELEON ) < 0 ) {
				syslog(LOG_ERR, 
				  "%s/%s: rpc SO_CHAMELEON setsockopt: %m",
					sep->se_service, sep->se_proto);
				(void)close(s);
				return -1;
			}
			if (turnon(s, SO_REUSEADDR) < 0) {
				syslog(LOG_ERR, 
				  "%s/%s: rpc SO_REUSEADDR setsockopt: %m",
					sep->se_service, sep->se_proto);
				(void)close(s);
				return -1;
			}
			if (turnon(s, SO_REUSEPORT) < 0) {
				syslog(LOG_ERR, 
				  "%s/%s: rpc SO_REUSEPORT setsockopt: %m",
					sep->se_service, sep->se_proto);
				(void)close(s);
				return -1;
			}
		}
	}
	addr.sin_family = AF_INET;
	addr.sin_port = 0;
	addr.sin_addr.s_addr = INADDR_ANY;

	if (havecipso && havemac && sep->se_port) {
		/* already registered port number, reuse it */
		addr.sin_port = htons(sep->se_port);
	}
	if (bind(s, &addr, sizeof(addr)) < 0) {
		syslog(LOG_ERR, "bind failed");
		(void)close(s);
		return -1;
	}
	if (havecipso && havemac && sep->se_port) {
		/* already registered port number, reuse it */
		goto skip_registration;
	}
	len = sizeof(struct sockaddr_in);

	if (getsockname(s, &addr, &len) != 0) {
		syslog(LOG_ERR, "getsockname failed");
		(void)close(s);
		return -1;
	}
	sep->se_prot = (sep->se_socktype == SOCK_DGRAM) ? IPPROTO_UDP
			: IPPROTO_TCP;
	sep->se_port = htons(addr.sin_port);
	rvp = &sep->se_vers;
	do {
		register u_long vers;

		for (vers = rvp->rv_low; vers <= rvp->rv_high; vers++) {
			if (debug) {
				fprintf(stderr, "%s: %d/%d/%d port = %d\n",
					sep->se_service, 
					sep->se_prog, vers,
					sep->se_prot, 
					sep->se_port);
			}
			if (!pmap_set(sep->se_prog, vers, sep->se_prot,
				      sep->se_port)) {
				if (pmap_unset(sep->se_prog, vers)) {
					if (pmap_set(sep->se_prog,
					    vers, sep->se_prot,
					    sep->se_port)) {
						registrations++;
					}
					if (sep->se_aliases) {
						pmap_reset_aliases(sep, 0);
					}
				}
			} else 
				registrations++;
		}
		rvp = rvp->rv_more;
	} while (rvp);

	if (registrations == 0) {
		/*
		 * all attempts to register this service failed!
		 * forget this port number
		 */
		syslog(LOG_ERR, "%s/%s: failed to register service",
			sep->se_service, sep->se_proto);

		sep->se_port = 0;
		close(s);
		return -1;
	}
skip_registration:
	if (sep->se_socktype == SOCK_STREAM)
		if (havecipso) {
			listen_qlen = IS_CHAMELEON(sep) ? 1 : LISTEN_QLEN;
		} else {
			listen_qlen = LISTEN_QLEN;
		}
		listen(s, listen_qlen);
	return s;
}

struct servtab *
enter(cp)
	struct servtab *cp;
{
	register struct servtab *sep;
	long omask;

	sep = tmalloc(struct servtab);
	*sep = *cp;
	sep->se_fd = -1;
	omask = sigblock(SIGBLOCK);
	sep->se_next = servtab;
	servtab = sep;
	sigsetmask(omask);
	return (sep);
}

struct	servtab serv;
#define MAX_LINE_LEN	256
char	line[MAX_LINE_LEN];
char	*sskip(), *skip(), *nextline();

setconfig()
{

	if (fconfig != NULL) {
		fseek(fconfig, 0L, L_SET);
		return (1);
	}
	fconfig = fopen(CONFIG, "r");
	return (fconfig != NULL);
}

endconfig()
{
	if (fconfig) {
		(void) fclose(fconfig);
		fconfig = NULL;
	}
}

struct servtab *
getconfigent()
{
  register struct servtab *sep = &serv;
  char *cp, *arg, *usrp, *tmpstr;
  int argc;
  static char TCPMUX_TOKEN[] = "tcpmux/";
#define MUX_LEN		(sizeof(TCPMUX_TOKEN)-1)

more:
  while ((cp = nextline(fconfig)) && (*cp == '#' || *cp == '\0'))
    ;
  if (cp == NULL)
    return ((struct servtab *)0);
  /*
   * clear the static buffer, since some fields (se_ctrladdr,
   * for example) don't get initialized here.
   */
  bzero((caddr_t)sep, sizeof *sep);
  arg = skip(&cp);
  if (cp == NULL) {
    /* got an empty line containing just blanks/tabs. */
    goto more;
  }
  if (strncmp(arg, TCPMUX_TOKEN, MUX_LEN) == 0) {
    char *c = arg + MUX_LEN;
    if (*c == '+') {
      sep->se_type = MUXPLUS_TYPE;
			c++;
    } else
      sep->se_type = MUX_TYPE;
		sep->se_service = mystrdup(c);
  } else {
    sep->se_service = mystrdup(arg);
		sep->se_type = NORM_TYPE;
  }
  arg = skip(&cp);
  if (strcmp(arg, "stream") == 0)
    sep->se_socktype = SOCK_STREAM;
  else if (strcmp(arg, "dgram") == 0)
    sep->se_socktype = SOCK_DGRAM;
  else if (strcmp(arg, "rdm") == 0)
    sep->se_socktype = SOCK_RDM;
  else if (strcmp(arg, "seqpacket") == 0)
    sep->se_socktype = SOCK_SEQPACKET;
  else if (strcmp(arg, "raw") == 0)
    sep->se_socktype = SOCK_RAW;
  else
    sep->se_socktype = -1;
  arg = sskip(&cp);
  if (strncmp(arg, "rpc/", 4) != 0) {
    sep->se_proto = mystrdup(arg);
  } else {
    register struct rpcent *rp;
    
    sep->se_proto = mystrdup(arg + 4);
    sep->se_type = RPC_TYPE;
    arg = strchr(sep->se_service, '/');
    if (arg == NULL) {
      syslog(LOG_ERR, "%s: missing rpc version for %s",
	     CONFIG, sep->se_service);
      goto more;
    }
    *arg++ = '\0';
    rp = getrpcbyname(sep->se_service);
    if (rp == NULL) {
      syslog(LOG_ERR, "%s: unknown rpc service %s", CONFIG,
	     sep->se_service);
      goto more;
    }
    sep->se_prog = rp->r_number;
		if (!getrpcvers(arg, &sep->se_vers)) {
		  syslog(LOG_ERR, "%s: malformed rpc version in %s",
			 CONFIG, sep->se_service);
		  goto more;
		}
	}
  arg = sskip(&cp);
	if (havecipso && havemac &&
	    (strcmp(arg, "wait/lc") == 0 || strcmp(arg, "nowait/lc") == 0))
	  sep->se_flags |= SE_LBLCOG;
  /*
   * If cipso/MAC, don't allow servers to be "wait" 
   * unless they're label cognizant.
   */
	if ((!havecipso || !havemac || 
	     IS_LBLCOG(sep)) && strncmp(arg, "wait", 4) == 0) 
	  sep->se_wait = 1;
  if (ISMUX(sep)) {
		/*
		 * Silently enforce "nowait" for TCPMUX services since
		 * they don't have an assigned port to listen on.
		 */
    sep->se_wait = 0;

    if (strcmp(sep->se_proto, "tcp")) {
      syslog(LOG_ERR, 
	     "%s: bad protocol for tcpmux service %s",
	     CONFIG, sep->se_service);
      goto more;
    }
    if (sep->se_socktype != SOCK_STREAM) {
      syslog(LOG_ERR, 
	     "%s: bad socket type for tcpmux service %s",
	     CONFIG, sep->se_service);
      goto more;
    }
  }
  sep->se_user = mystrdup(sskip(&cp));
  if ((usrp = strstr(sep->se_user, "/rcv")) != 0) {
    *usrp = 0;
    if (havecipso) 
      sep->se_flags |= SE_RCVUID;
  }
  tmpstr = sskip(&cp);
  if (*tmpstr == '?') {
    sep->se_flags |= SE_MISSING_OK;
    sep->se_server = mystrdup(tmpstr+1);
  } else
    sep->se_server = mystrdup(tmpstr);
  if (strcmp(sep->se_server, "internal") == 0) {
    register struct biltin *bi;
    
    for (bi = biltins; bi->bi_service; bi++)
      if (bi->bi_socktype == sep->se_socktype &&
	  strcmp(bi->bi_service, sep->se_service) == 0)
	break;
    if (bi->bi_service == 0) {
      syslog(LOG_ERR, "internal service %s unknown",
	     sep->se_service);
      goto more;
    }
    sep->se_bi = bi;
    sep->se_wait = bi->bi_wait;
  } else
    sep->se_bi = NULL;
  if (havecipso || havemac) {
    /* additional sanity checks */
    if (sep->se_socktype == SOCK_STREAM && 
	strcmp(sep->se_proto, "tcp") ||
	sep->se_socktype == SOCK_DGRAM && 
	strcmp(sep->se_proto, "udp") ||
	sep->se_type     == RPC_TYPE && sep->se_bi) {
      syslog(LOG_ERR, 
	     "%s: bad configuration combination for service %s",
	     CONFIG, sep->se_service);
			goto more;
    }
    /* 
     * If server is not label-cognizant, it should have
     * the chameleon flag unless it is non-RPC, stream
     */
		if (!IS_LBLCOG(sep) && !sep->se_bi && 
		    (sep->se_type == RPC_TYPE || sep->se_socktype == SOCK_DGRAM)) {
		  sep->se_flags |= SE_CHAMELEON;
		}
	}
  argc = 0;
  for (arg = skip(&cp); cp; arg = skip(&cp))
		if (argc < MAXARGV)
		  sep->se_argv[argc++] = mystrdup(arg);
  while (argc <= MAXARGV)
    sep->se_argv[argc++] = NULL;
  sep->se_aliases = sep->se_nextalias = NULL;
  return (sep);
}

int
getrpcvers(buf, rvp)
	char *buf;
	struct rpcvers *rvp;
{
	char *cp;

	rvp->rv_low = strtoul(buf, &cp, 0);
	if (cp == buf)
		return 0;
	if (*cp == '-') {
		buf = cp + 1;
		rvp->rv_high = strtoul(buf, &cp, 0);
		if (cp == buf)
			return 0;
	} else
		rvp->rv_high = rvp->rv_low;
	if (*cp == ',') {
		rvp->rv_more = tmalloc(struct rpcvers);
		if (!getrpcvers(cp + 1, rvp->rv_more)) {
			free((char *) rvp->rv_more);
			return 0;
		}
		return 1;
	}
	rvp->rv_more = 0;
	return (*cp == '\0');
}

freeconfig(cp)
	register struct servtab *cp;
{
	int i;

	if (cp->se_service && !ISRPC(cp))
		free(cp->se_service);
	if (cp->se_proto)
		free(cp->se_proto);
	if (cp->se_user)
		free(cp->se_user);
	if (cp->se_server)
		free(cp->se_server);
	for (i = 0; i < MAXARGV; i++)
		if (cp->se_argv[i])
			free(cp->se_argv[i]);
	if (ISRPC(cp))
		freerpcvers(cp);
}

freerpcvers(sep)
	struct servtab *sep;
{
	struct rpcvers *rvp, *more;

	for (rvp = sep->se_vers.rv_more; rvp; rvp = more) {
		more = rvp->rv_more;
		free((char *) rvp);
	}
}

/*
 * Safe skip - if skip returns null, log a syntax error in the
 * configuration file and exit.
 */
char *
sskip(cpp)
	char **cpp;
{
	register char *cp;

	cp = skip(cpp);
	if (cp == NULL) {
		syslog(LOG_ERR, "%s: syntax error", CONFIG);
		exit(-1);
	}
	return cp;
}

char *
skip(cpp)
	char **cpp;
{
	register char *cp = *cpp;
	char *start;

again:
	while (*cp == ' ' || *cp == '\t')
		cp++;
	if (*cp == '\0') {
		int c;

		c = getc(fconfig);
		(void) ungetc(c, fconfig);
		if (c == ' ' || c == '\t')
			if (cp = nextline(fconfig))
				goto again;
		*cpp = (char *)0;
		return ((char *)0);
	}
	start = cp;
	while (*cp && *cp != ' ' && *cp != '\t')
		cp++;
	if (*cp != '\0')
		*cp++ = '\0';
	*cpp = cp;
	return (start);
}

char *
nextline(fd)
	FILE *fd;
{
	char *cp;

	if (fgets(line, sizeof (line), fd) == NULL)
		return ((char *)0);
	cp = index(line, '\n');
	if (cp)
		*cp = '\0';
	return (line);
}

char *
emalloc(size)
	int size;
{
	char *new;

	new = malloc(size);
	if (new == 0) {
		syslog(LOG_ERR, "Out of memory.");
		exit(-1);
	}
	return new;
}

/*
 * Internet services provided internally by inetd:
 */
#define	BUFSIZE	(16 * 1024)

/* ARGSUSED */
echo_stream(s, sep)		/* Echo service -- echo data back */
	int s;
	struct servtab *sep;
{
	char buffer[BUFSIZE];
	int i;

	while ((i = read(s, buffer, sizeof(buffer))) > 0 &&
	    write(s, buffer, i) > 0)
		;
	exit(0);
}

/* ARGSUSED */
echo_dg(s, sep)			/* Echo service -- echo data back */
	int s;
	struct servtab *sep;
{
	char buffer[BUFSIZE];
	int i, size;
	struct sockaddr sa;

	size = sizeof(sa);
	if ((i = recvfrom(s, buffer, sizeof(buffer), 0, &sa, &size)) < 0)
		return;
	(void) sendto(s, buffer, i, 0, &sa, sizeof(sa));
}

/* ARGSUSED */
discard_stream(s, sep)		/* Discard service -- ignore data */
	int s;
	struct servtab *sep;
{
	char buffer[BUFSIZE];
	errno = 0;

	while (1) {
		while (read(s, buffer, sizeof(buffer)) > 0)
			;
		if (errno != EINTR)
			break;
	}
	exit(0);
}

/* ARGSUSED */
discard_dg(s, sep)		/* Discard service -- ignore data */
	int s;
	struct servtab *sep;
{
	char buffer[BUFSIZE];

	(void) read(s, buffer, sizeof(buffer));
}

#include <ctype.h>
#define LINESIZ 72
char ring[128];
char *endring;

initring()
{
	register int i;

	endring = ring;

	for (i = 0; i <= 128; ++i)
		if (isprint(i))
			*endring++ = i;
}

/* ARGSUSED */
chargen_stream(s, sep)		/* Character generator */
	int s;
	struct servtab *sep;
{
	register char *rs;
	int len;
	char text[LINESIZ+2];


	if (!endring) {
		initring();
		rs = ring;
	}

	text[LINESIZ] = '\r';
	text[LINESIZ + 1] = '\n';
	for (rs = ring;;) {
		if ((len = endring - rs) >= LINESIZ)
			bcopy(rs, text, LINESIZ);
		else {
			bcopy(rs, text, len);
			bcopy(ring, text + len, LINESIZ - len);
		}
		if (++rs == endring)
			rs = ring;
		if (write(s, text, sizeof(text)) != sizeof(text))
			break;
	}
	exit(0);
}

/* ARGSUSED */
chargen_dg(s, sep)		/* Character generator */
	int s;
	struct servtab *sep;
{
	struct sockaddr sa;
	static char *rs;
	int len, size;
	char text[LINESIZ+2];

	if (endring == 0) {
		initring();
		rs = ring;
	}

	size = sizeof(sa);
	if (recvfrom(s, text, sizeof(text), 0, &sa, &size) < 0)
		return;

	if ((len = endring - rs) >= LINESIZ)
		bcopy(rs, text, LINESIZ);
	else {
		bcopy(rs, text, len);
		bcopy(ring, text + len, LINESIZ - len);
	}
	if (++rs == endring)
		rs = ring;
	text[LINESIZ] = '\r';
	text[LINESIZ + 1] = '\n';
	(void) sendto(s, text, sizeof(text), 0, &sa, sizeof(sa));
}

/*
 * Return a machine readable date and time, in the form of the
 * number of seconds since midnight, Jan 1, 1900.  Since gettimeofday
 * returns the number of seconds since midnight, Jan 1, 1970,
 * we must add 2208988800 seconds to this figure to make up for
 * some seventy years Bell Labs was asleep.
 */

long
machtime()
{
	struct timeval tv;

	if (gettimeofday(&tv, (struct timezone *)0) < 0) {
		if (debug)
			fprintf(stderr, "Unable to get time of day\n");
		return (0L);
	}
	return (htonl((long)tv.tv_sec + 2208988800));
}

/* ARGSUSED */
machtime_stream(s, sep)
	int s;
	struct servtab *sep;
{
	long result;

	result = machtime();
	(void) write(s, (char *) &result, sizeof(result));
}

/* ARGSUSED */
machtime_dg(s, sep)
	int s;
	struct servtab *sep;
{
	long result;
	struct sockaddr sa;
	int size;

	size = sizeof(sa);
	if (recvfrom(s, (char *)&result, sizeof(result), 0, &sa, &size) < 0)
		return;
	result = machtime();
	(void) sendto(s, (char *) &result, sizeof(result), 0, &sa, sizeof(sa));
}

/* ARGSUSED */
daytime_stream(s, sep)		/* Return human-readable time of day */
	int s;
	struct servtab *sep;
{
	char buffer[256];
	time_t clock;

	clock = time((time_t *) 0);

	(void) sprintf(buffer, "%.24s\r\n", ctime(&clock));
	(void) write(s, buffer, strlen(buffer));
}

/* ARGSUSED */
daytime_dg(s, sep)		/* Return human-readable time of day */
	int s;
	struct servtab *sep;
{
	char buffer[256];
	time_t clock;
	struct sockaddr sa;
	int size;

	clock = time((time_t *) 0);

	size = sizeof(sa);
	if (recvfrom(s, buffer, sizeof(buffer), 0, &sa, &size) < 0)
		return;
	(void) sprintf(buffer, "%.24s\r\n", ctime(&clock));
	(void) sendto(s, buffer, strlen(buffer), 0, &sa, sizeof(sa));
}

/*
 * print_service:
 *	Dump relevant information to stderr
 */
print_service(action, sep)
	char *action;
	struct servtab *sep;
{
	fprintf(stderr,
	    "%s: %s proto=%s, wait=%d, user=%s builtin=%x server=%s\n",
	    action, sep->se_service, sep->se_proto,
	    sep->se_wait, sep->se_user, (int)sep->se_bi, sep->se_server);
}

/*
 *  Based on TCPMUX.C by Mark K. Lottor November 1988
 *  sri-nic::ps:<mkl>tcpmux.c
 */


static int		/* # of characters upto \r,\n or \0 */
getline(fd, buf, len)
	int fd;
	char *buf;
	int len;
{
	int count = 0, n;

	do {
		n = read(fd, buf, len-count);
		if (n == 0)
			return count;
		if (n < 0)
			return (-1);
		while (--n >= 0) {
			if (*buf == '\r' || *buf == '\n' || *buf == '\0')
				return count;
			count++;
			buf++;
		}
	} while (count < len);
	return (count);
}

#define MAX_SERV_LEN	(MAX_LINE_LEN+2)	/* 2 bytes for \r\n */

#define strwrite(fd, buf)	(void) write(fd, buf, sizeof(buf)-1)

struct servtab *
tcpmux(s)
	int s;
{
	register struct servtab *sep;
	char service[MAX_SERV_LEN+1];
	int len;

	/* Get requested service name */
	if ((len = getline(s, service, MAX_SERV_LEN)) < 0) {
	    strwrite(s, "-Error reading service name\r\n");
	    return(NULL);
	}
	service[len] = '\0';

	if (debug)
	    fprintf(stderr, "tcpmux: someone wants %s\n", service);

	/*
	 * Help is a required command, and lists available services,
	 * one per line.
	 */
	if (!strcasecmp(service,"help")) {
	    for (sep = servtab; sep; sep = sep->se_next) {
		if (!ISMUX(sep))
		    continue;
		(void) write(s, sep->se_service, strlen(sep->se_service));
		strwrite(s, "\r\n");
	    }
	    return(NULL);
	}

	/* Try matching a service in inetd.conf with the request */
	for (sep = servtab; sep; sep = sep->se_next) {
	    if (!ISMUX(sep))
		continue;
	    if (!strcasecmp(service,sep->se_service)) {
		if (ISMUXPLUS(sep)) {
		    strwrite(s, "+Go\r\n");
		}
		return(sep);
	    }
	}
	strwrite(s, "-Service not available\r\n");
	return(NULL);
}

/* Check user clearance for current process label.  */
void 
check_clearance(struct clearance *clp, struct servtab *sep)
{
	mac_label proc_label;
	caddr_t labelname = NULL;

	bzero(&proc_label, sizeof proc_label);
	if (havemac) {
		if (sgi_getplabel(&proc_label) < 0 ||
		    mac_clearedlbl(clp, &proc_label) != MAC_CLEARED) {
			labelname = mac_labeltostr(&proc_label, 1);
			syslog(LOG_ERR|LOG_AUTH, 
			       "%s not cleared for %s at label %.40s",
			       clp->cl_name, sep->se_service, 
			       labelname ? labelname : "** INVALID **");
			(void)ia_audit("inetd", clp->cl_name, 0, "Not cleared.");
 			_exit(1);
		}
	}
 	(void)ia_audit("inetd", clp->cl_name, 1, "Cleared for label.");
}
